function [minx,minf,x_t,f_t,S_k] = FW(x_0, S_0, alpha_0, A, b, fun_optim, opts) 
% [x_t, f_t] = FW(x_0, S_0, alpha_0, A, fun_optim, opts) 
% x_0 : variable initialization 
% S_0 : active set initialization 
% alpha_0 : weight initialization 
% A : the matrix in the cost function of the QP
% fun_optim : the function to solve the related LP:
% min_z <z, grad(x_t) >
% for images : take the min of the integer solutions (eg z=(1,0,0) )
% for videos : shortest path algorithm


N           = size(x_0, 1);
max_step    = 1;
it           = 1;
minf = 0;
minx = [];
% init:

x_t         = x_0;
S_k         = S_0;
alpha_k     = alpha_0;

cost_fun =@(y) .5 * y' *A * y + b' * y;
% optimization: 

while it <= opts.Tmax
  it = it + 1; 

  % gradient:

  grad        = A * x_t + b;

  % cost function:

  f_t = cost_fun(x_t);

  % normal update:
  
  y_T      = fun_optim( grad );
  d_T      = y_T - x_t;

  if it==2 || cost_fun(y_T) < minf
      minx = y_T;
      minf = cost_fun(y_T)
  end

  % duality gap:

  gap = - d_T' * grad;

  if opts.verbose
      fprintf('it = %d -  f = %f - gap=%f\n', it, f_t, gap);
  end

  if gap < opts.TOL
    fprintf('end of FW: reach small duality gap (gap=%f)\n', gap);
    break
  end 
  
  % away update:

  if ~isempty(S_k) 
    id_A   = away_step(grad, S_k);
    y_A    = S_k(:, id_A);
    d_A    = x_t - y_A;
    alpha_max = alpha_k(id_A);
  end

  % choose direction (between normal and away):

  if isempty(S_k) || - gap <= d_A' * grad
    is_aw = false;
    y = y_T;
    d = d_T; 
    max_step = 1;
  else
    is_aw = true;
    y = y_A;
    d = d_A;
    max_step = alpha_max / (1 - alpha_max);
  end
  
  % line search:

  step = -  (grad' * d) / ( d' * A * d );
  step = max(0, min(step, max_step ));  

  if step < opts.TOL
    fprintf('end of FW: step is too small (step=%f)\n', step)
    break
  end

  [S_k, alpha_k] = update_active_set(S_k, alpha_k, y, step, max_step, is_aw);

  x_t = x_t + step * d; 
 
end

y = fun_optim(- x_t);
fy = cost_fun(y)

if fy < minf
  minf = fy;
  minx = y;
end

function id = away_step(grad, S)
    s = S' * grad;
    [~,id] = max(s);
    id = id(1);
end


function [S_k, alpha_k] = update_active_set(S_k, alpha_k, v_k, gamma_k, gamma_max, is_aw)
    S_0 = S_k;
    a_0 = alpha_k;

  if ~is_aw % normal step:
     if isempty(S_k) || gamma_k == gamma_max
         S_k = v_k;
         alpha_k = 1;
     else 
         [is_in, id] =ismember(S_k', v_k', 'rows');
         is_in = sum(is_in)~=0;
         id = id == 1;

         alpha_k = alpha_k * ( 1 - gamma_k); 

         if is_in
            alpha_k(id) = alpha_k(id) + gamma_k;
         else
            S_k = [ S_k , v_k];
            alpha_k = [alpha_k , gamma_k];
         end
     end
  
  else % away step
  
     [~, id] =ismember(S_k', v_k', 'rows');
     id = id ==1;
     
     alpha_k = alpha_k * ( 1 + gamma_k); 

     if gamma_k == gamma_max % drop step
         S_k(:,id) = [];
         alpha_k(id)=[];
     else
         alpha_k(id) = alpha_k(id) - gamma_k;
     end

  end
 
%  S_k = S_k(:, alpha_k>opts.eps);
%  alpha_k = alpha_k(alpha_k>opts.eps);

end

end
